@@ -3,9 +3,10 @@
  * \brief Computes an approximate solutions of linear equations A*X=B or A'*X=B
  *
  * <pre>
- * -- SuperLU routine (version 4.1) --
+ * -- SuperLU routine (version 4.2) --
  * Lawrence Berkeley National Laboratory.
  * November, 2010
+ * August, 2011
  * </pre>
  */
 #include "slu_ddefs.h"
@@ -219,7 +220,9 @@
  *            off-diagonal entries of modulus at most 1. P is a permutation
  *            obtained from MC64.
  *            If MC64 fails, dgsequ() is used to equilibrate the system,
- *            and A is scaled as above, there is no permutation involved.
+ *            and A is scaled as above, but no permutation is involved.
+ *            On exit, A is restored to the orginal row numbering, so
+ *            Dr*A*Dc is returned.
  *
  * perm_c  (input/output) int*
  *	   If A->Stype = SLU_NC, Column permutation vector of size A->ncol,
@@ -237,8 +240,8 @@
  * perm_r  (input/output) int*
  *	   If A->Stype = SLU_NC, row permutation vector of size A->nrow, 
  *	   which defines the permutation matrix Pr, and is determined
- *	   by partial pivoting.  perm_r[i] = j means row i of A is in 
- *	   position j in Pr*A.
+ *	   by MC64 first then followed by partial pivoting.
+ *         perm_r[i] = j means row i of A is in position j in Pr*A.
  *
  *	   If A->Stype = SLU_NR, permutation vector of size A->ncol, which
  *	   determines permutation of rows of transpose(A)
@@ -275,7 +278,7 @@
  *	   If equed = 'N' or 'C', R is not accessed.
  *	   If options->Fact = FACTORED, R is an input argument,
  *	       otherwise, R is output.
- *	   If options->zFact = FACTORED and equed = 'R' or 'B', each element
+ *	   If options->Fact = FACTORED and equed = 'R' or 'B', each element
  *	       of R must be positive.
  *
  * C	   (input/output) double*, dimension (A->ncol)
@@ -336,11 +339,6 @@
  *		 if options->Trans = TRANS or CONJ and equed = 'R' of 'B',
  *		    B is overwritten by diag(R)*B.
  *
- *         If options->RowPerm = LargeDiag, MC64 is used to scale and permute
- *            the matrix A to an I-matrix. Then, in addition to the scaling
- *            above, B is further permuted by P*B if options->Trans = NOTRANS,
- *            where P is obtained from MC64.
- *
  * X	   (output) SuperMatrix*
  *	   X has types: Stype = SLU_DN, Dtype = SLU_D, Mtype = SLU_GE.
  *	   If info = 0 or info = A->ncol+1, X contains the solution matrix
@@ -398,12 +396,12 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
        SuperMatrix *L, SuperMatrix *U, void *work, int lwork,
        SuperMatrix *B, SuperMatrix *X,
        double *recip_pivot_growth, double *rcond,
-       mem_usage_t *mem_usage, SuperLUStat_t *stat, int *info)
+       GlobalLU_t *Glu, mem_usage_t *mem_usage, SuperLUStat_t *stat, int *info)
 {
 
     DNformat  *Bstore, *Xstore;
     double    *Bmat, *Xmat;
-    int       ldb, ldx, nrhs;
+    int       ldb, ldx, nrhs, n;
     SuperMatrix *AA;/* A in SLU_NC format used by the factorization routine.*/
     SuperMatrix AC; /* Matrix postmultiplied by Pc */
     int       colequ, equil, nofact, notran, rowequ, permc_spec, mc64;
@@ -416,7 +414,7 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
     double    t0;      /* temporary time */
     double    *utime;
 
-    int *perm = NULL;
+    int *perm = NULL; /* permutation returned from MC64 */
 
     /* External functions */
     extern double dlangs(char *, SuperMatrix *);
@@ -428,6 +426,7 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
     ldb    = Bstore->lda;
     ldx    = Xstore->lda;
     nrhs   = B->ncol;
+    n      = B->nrow;
 
     *info = 0;
     nofact = (options->Fact != FACTORED);
@@ -439,24 +438,26 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
 	rowequ = FALSE;
 	colequ = FALSE;
     } else {
-	rowequ = lsame_(equed, "R") || lsame_(equed, "B");
-	colequ = lsame_(equed, "C") || lsame_(equed, "B");
-	smlnum = dlamch_("Safe minimum");
+	rowequ = strncmp(equed, "R", 1)==0 || strncmp(equed, "B", 1)==0;
+	colequ = strncmp(equed, "C", 1)==0 || strncmp(equed, "B", 1)==0;
+	smlnum = dmach("Safe minimum");  /* lamch_("Safe minimum"); */
 	bignum = 1. / smlnum;
     }
 
     /* Test the input parameters */
-    if (!nofact && options->Fact != DOFACT && options->Fact != SamePattern &&
+    if (options->Fact != DOFACT && options->Fact != SamePattern &&
 	options->Fact != SamePattern_SameRowPerm &&
-	!notran && options->Trans != TRANS && options->Trans != CONJ &&
-	!equil && options->Equil != NO)
+	options->Fact != FACTORED &&
+	options->Trans != NOTRANS && options->Trans != TRANS && 
+	options->Trans != CONJ &&
+	options->Equil != NO && options->Equil != YES)
 	*info = -1;
     else if ( A->nrow != A->ncol || A->nrow < 0 ||
 	      (A->Stype != SLU_NC && A->Stype != SLU_NR) ||
 	      A->Dtype != SLU_D || A->Mtype != SLU_GE )
 	*info = -2;
-    else if (options->Fact == FACTORED &&
-	     !(rowequ || colequ || lsame_(equed, "N")))
+    else if ( options->Fact == FACTORED &&
+	     !(rowequ || colequ || strncmp(equed, "N", 1)==0) )
 	*info = -6;
     else {
 	if (rowequ) {
@@ -498,7 +499,7 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
     }
     if (*info != 0) {
 	i = -(*info);
-	xerbla_("dgsisx", &i);
+	input_error("dgsisx", &i);
 	return;
     }
 
@@ -535,38 +536,46 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
 	int *colptr = Astore->colptr;
 	int *rowind = Astore->rowind;
 	double *nzval = (double *)Astore->nzval;
-	int n = AA->nrow;
 
 	if ( mc64 ) {
-	    *equed = 'B';
-    	    /*rowequ = colequ = 1;*/
 	    t0 = SuperLU_timer_();
 	    if ((perm = intMalloc(n)) == NULL)
 		ABORT("SUPERLU_MALLOC fails for perm[]");
 
 	    info1 = dldperm(5, n, nnz, colptr, rowind, nzval, perm, R, C);
 
-	    if (info1 > 0) { /* MC64 fails, call dgsequ() later */
+	    if (info1 != 0) { /* MC64 fails, call dgsequ() later */
 		mc64 = 0;
 		SUPERLU_FREE(perm);
 		perm = NULL;
 	    } else {
-	        rowequ = colequ = 1;
-		for (i = 0; i < n; i++) {
-		    R[i] = exp(R[i]);
-		    C[i] = exp(C[i]);
-		}
-		/* permute and scale the matrix */
+	        if ( equil ) {
+	            rowequ = colequ = 1;
+		    for (i = 0; i < n; i++) {
+		        R[i] = exp(R[i]);
+		        C[i] = exp(C[i]);
+		    }
+		    /* scale the matrix */
+		    for (j = 0; j < n; j++) {
+		        for (i = colptr[j]; i < colptr[j + 1]; i++) {
+			    nzval[i] *= R[rowind[i]] * C[j];
+		        }
+		    }
+	            *equed = 'B';
+                }
+
+                /* permute the matrix */
 		for (j = 0; j < n; j++) {
 		    for (i = colptr[j]; i < colptr[j + 1]; i++) {
-			nzval[i] *= R[rowind[i]] * C[j];
+			/*nzval[i] *= R[rowind[i]] * C[j];*/
 			rowind[i] = perm[rowind[i]];
 		    }
 		}
 	    }
 	    utime[EQUIL] = SuperLU_timer_() - t0;
 	}
-	if ( !mc64 & equil ) {
+
+	if ( !mc64 & equil ) { /* Only perform equilibration, no row perm */
 	    t0 = SuperLU_timer_();
 	    /* Compute row and column scalings to equilibrate the matrix A. */
 	    dgsequ(AA, R, C, &rowcnd, &colcnd, &amax, &info1);
@@ -574,8 +583,8 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
 	    if ( info1 == 0 ) {
 		/* Equilibrate matrix A. */
 		dlaqgs(AA, R, C, rowcnd, colcnd, amax, equed);
-		rowequ = lsame_(equed, "R") || lsame_(equed, "B");
-		colequ = lsame_(equed, "C") || lsame_(equed, "B");
+		rowequ = strncmp(equed, "R", 1)==0 || strncmp(equed, "B", 1)==0;
+		colequ = strncmp(equed, "C", 1)==0 || strncmp(equed, "B", 1)==0;
 	    }
 	    utime[EQUIL] = SuperLU_timer_() - t0;
 	}
@@ -605,13 +614,33 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
 	/* Compute the LU factorization of A*Pc. */
 	t0 = SuperLU_timer_();
 	dgsitrf(options, &AC, relax, panel_size, etree, work, lwork,
-                perm_c, perm_r, L, U, stat, info);
+                perm_c, perm_r, L, U, Glu, stat, info);
 	utime[FACT] = SuperLU_timer_() - t0;
 
 	if ( lwork == -1 ) {
 	    mem_usage->total_needed = *info - A->ncol;
 	    return;
 	}
+
+	if ( mc64 ) { /* Fold MC64's perm[] into perm_r[]. */
+	    NCformat *Astore = AA->Store;
+	    int nnz = Astore->nnz, *rowind = Astore->rowind;
+	    int *perm_tmp, *iperm;
+	    if ((perm_tmp = intMalloc(2*n)) == NULL)
+		ABORT("SUPERLU_MALLOC fails for perm_tmp[]");
+	    iperm = perm_tmp + n;
+	    for (i = 0; i < n; ++i) perm_tmp[i] = perm_r[perm[i]];
+	    for (i = 0; i < n; ++i) {
+		perm_r[i] = perm_tmp[i];
+		iperm[perm[i]] = i;
+	    }
+
+	    /* Restore A's original row indices. */
+	    for (i = 0; i < nnz; ++i) rowind[i] = iperm[rowind[i]];
+
+	    SUPERLU_FREE(perm); /* MC64 permutation */
+	    SUPERLU_FREE(perm_tmp);
+	}
     }
 
     if ( options->PivotGrowth ) {
@@ -635,12 +664,7 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
     }
 
     if ( nrhs > 0 ) { /* Solve the system */
-        double *tmp, *rhs_work;
-        int n = A->nrow;
-        if ( mc64 ) {
-	    if ((tmp = doubleMalloc(n)) == NULL)
-		ABORT("SUPERLU_MALLOC fails for tmp[]");
-        }
+        double *rhs_work;
 
 	/* Scale and permute the right-hand side if equilibration
            and permutation from MC64 were performed. */
@@ -650,13 +674,6 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
 		    for (i = 0; i < n; ++i)
 		        Bmat[i + j*ldb] *= R[i];
 	    }
-	    if ( mc64 ) {
-		for (j = 0; j < nrhs; ++j) {
-                   rhs_work = &Bmat[j*ldb];
-  	           for (i = 0; i < n; i++) tmp[perm[i]] = rhs_work[i];
-	           for (i = 0; i < n; i++) rhs_work[i] = tmp[i];
-                }
-	    }
 	} else if ( colequ ) {
 	    for (j = 0; j < nrhs; ++j)
 		for (i = 0; i < n; ++i) {
@@ -684,33 +701,21 @@ dgsisx(superlu_options_t *options, SuperMatrix *A, int *perm_c, int *perm_r,
 	    }
 	} else { /* transposed system */
 	    if ( rowequ ) {
-		if ( mc64 ) {
-		    for (j = 0; j < nrhs; j++) {
-			for (i = 0; i < n; i++)
-			    tmp[i] = Xmat[i + j * ldx]; /*dcopy*/
-			for (i = 0; i < n; i++)
-			    Xmat[i + j * ldx] = R[i] * tmp[perm[i]];
-		    }
-		} else {
-		    for (j = 0; j < nrhs; ++j)
-			for (i = 0; i < A->nrow; ++i) {
-              	            Xmat[i + j*ldx] *= R[i];
-                        }
-		}
+	        for (j = 0; j < nrhs; ++j)
+		    for (i = 0; i < A->nrow; ++i) {
+              	        Xmat[i + j*ldx] *= R[i];
+                    }
 	    }
 	}
 
-        if ( mc64 ) SUPERLU_FREE(tmp);
-
     } /* end if nrhs > 0 */
 
     if ( options->ConditionNumber ) {
-	/* Set INFO = A->ncol+1 if the matrix is singular to working precision. */
-	if ( *rcond < dlamch_("E") && *info == 0) *info = A->ncol + 1;
+	/* The matrix is singular to working precision. */
+	/* if ( *rcond < dlamch_("E") && *info == 0) *info = A->ncol + 1; */
+	if ( *rcond < dmach("E") && *info == 0) *info = A->ncol + 1;
     }
 
-    if (perm) SUPERLU_FREE(perm);
-
     if ( nofact ) {
 	ilu_dQuerySpace(L, U, mem_usage);
 	Destroy_CompCol_Permuted(&AC);
